// Path information for API requests.
const baseUrl = window.location.origin;
const vpnControlPath = 'local/orchid_vpn_connector/vpn-control.cgi';
const vpnConfigPath = 'local/orchid_vpn_connector/vpn-config.cgi';
const vpnIpPath = 'local/orchid_vpn_connector/vpn-ip.cgi';
const vpnLogPath ='axis-cgi/admin/systemlog.cgi?appname=orchid_vpn_connector';
const vapixServicesPath = 'vapix/services';

// SOAP Request constants
const soapHeader = '<SOAP-ENV:Envelope xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:tds="http://www.onvif.org/ver10/device/wsdl" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:onvif="http://www.onvif.org/ver10/schema" xmlns:tt="http://www.onvif.org/ver10/schema" xmlns:SOAP-ENV="http://www.w3.org/2003/05/soap-envelope"><SOAP-ENV:Body>';
const soapFooter = '</SOAP-ENV:Body></SOAP-ENV:Envelope>';
const soapGetUsersRequestBody = '<tds:GetUsers xmlns="http://www.onvif.org/ver10/device/wsdl"></tds:GetUsers>';
const soapGetUsersRequest = soapHeader + soapGetUsersRequestBody + soapFooter;
const soapUpdateUserRequestBody = (isUpdate, password) => {
    let command = isUpdate ? 'SetUser' : 'CreateUsers';
    return `<tds:${command} xmlns="http://www.onvif.org/ver10/device/wsdl"><User><tt:Username>alto</tt:Username><tt:Password>${password}</tt:Password><tt:UserLevel>Administrator</tt:UserLevel></User></tds:${command}>`;
}
const soapUpdateUserRequest = (isUpdate, password) => {
    return soapHeader + soapUpdateUserRequestBody(isUpdate, password) + soapFooter;
};

const blueCheck = `<svg class="w-6 h-6 pt-1 fill-current text-blue-500" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 512 512"><!-- Font Awesome Free 5.15.3 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) --><path d="M173.898 439.404l-166.4-166.4c-9.997-9.997-9.997-26.206 0-36.204l36.203-36.204c9.997-9.998 26.207-9.998 36.204 0L192 312.69 432.095 72.596c9.997-9.997 26.207-9.997 36.204 0l36.203 36.204c9.997 9.997 9.997 26.206 0 36.204l-294.4 294.401c-9.998 9.997-26.207 9.997-36.204-.001z"/></svg>`;

const redX = `<svg class="w-6 h-6 pt-1 fill-current text-red-500" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 352 512"><!-- Font Awesome Free 5.15.3 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) --><path d="M242.72 256l100.07-100.07c12.28-12.28 12.28-32.19 0-44.48l-22.24-22.24c-12.28-12.28-32.19-12.28-44.48 0L176 189.28 75.93 89.21c-12.28-12.28-32.19-12.28-44.48 0L9.21 111.45c-12.28 12.28-12.28 32.19 0 44.48L109.28 256 9.21 356.07c-12.28 12.28-12.28 32.19 0 44.48l22.24 22.24c12.28 12.28 32.2 12.28 44.48 0L176 322.72l100.07 100.07c12.28 12.28 32.2 12.28 44.48 0l22.24-22.24c12.28-12.28 12.28-32.19 0-44.48L242.72 256z"/></svg>`

const spinner = `<svg class="w-6 h-6 pt-1 animate-spin" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 512 512"><!-- Font Awesome Free 5.15.3 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) --><path d="M304 48c0 26.51-21.49 48-48 48s-48-21.49-48-48 21.49-48 48-48 48 21.49 48 48zm-48 368c-26.51 0-48 21.49-48 48s21.49 48 48 48 48-21.49 48-48-21.49-48-48-48zm208-208c-26.51 0-48 21.49-48 48s21.49 48 48 48 48-21.49 48-48-21.49-48-48-48zM96 256c0-26.51-21.49-48-48-48S0 229.49 0 256s21.49 48 48 48 48-21.49 48-48zm12.922 99.078c-26.51 0-48 21.49-48 48s21.49 48 48 48 48-21.49 48-48c0-26.509-21.491-48-48-48zm294.156 0c-26.51 0-48 21.49-48 48s21.49 48 48 48 48-21.49 48-48c0-26.509-21.49-48-48-48zM108.922 60.922c-26.51 0-48 21.49-48 48s21.49 48 48 48 48-21.49 48-48-21.491-48-48-48z"/></svg>`

// Retrieve the expected DOM elements.
const elements = getDOMElements();

// Indicate if the camera contains an ONVIF Administrator user named 'alto'.
let onvifAltoUserExists = false;

// Interval for retrieving and rendering the logs.
//let logInterval = setInterval(updateLogs, 1000);

/**
 * Redirect a click to the file upload input.
 */
function clickFileInput() {
    elements.vpnUpload.click();
}

/**
 * Call endpoint to execute VPN controls and status.
 *
 * @param {string} action - VPN control action [start, stop, restart, status]
 * @returns {Promise} - A Promise that resolves with the response body message.
 */
function fetchVPNControl(action) {
    let url = new URL(vpnControlPath, baseUrl);
    url.searchParams.set('action', action);
    return fetch(url.toString())
}

function fetchTunnelAddress() {
    let url = new URL(vpnIpPath, baseUrl);
    return fetch(url.toString())
        .then(response => {
            if (response.ok) return response.text();
            return '';
        });
}

/**
 * Retrieve the current VPN status and update the view.
 *
 * @returns {Promise} - A Promise that resolves once the view has been updated.
 */
function updateVPNStatus() {
    // Indicate if the ACAP app is running.
    let isAppRunning = false;

    // Indicate if a VPN configuration has been applied to the camera.
    let isVpnConfigured = false;

    // Indicate if the VPN service is running.
    let isVpnRunning = false;

    let tunnelIp = '';


    return fetchVPNControl('status')
        .then(response => {
            isAppRunning = response.status < 500;
            isVpnConfigured = response.ok;

            return response.text();
        })
        .then(text => {
            isVpnRunning = text === 'VPN running\n';

            if (isVpnRunning) return fetchTunnelAddress();
            return '';
        })
        .then(address => {
            tunnelIp = address;
        })
        .catch(error => { console.error(error); })
        .finally(() => {
            updateVPNSection(isAppRunning, isVpnConfigured, isVpnRunning, tunnelIp);
        });
}

/**
 * Change an overall status DOM element to a new visual state (checkmark or x).
 *
 * @param {HTMLElement} statusElement - DOM element for the overall status to modify.
 * @param {bool} isSuccess - Indicate the status state.
 */
function updateOverallStatusElement(statusElement, isSuccess) {
    if (isSuccess === true) {
        statusElement.innerHTML = blueCheck;
    } else if (isSuccess === false) {
        statusElement.innerHTML = redX;
    } else {
        statusElement.innerHTML = spinner;
    }
}

/**
 * Update the VPN section of the DOM based on the new VPN state info.
 *
 * @param {bool} isAppRunning - TRUE if the ACAP application is running.
 * @param {bool} isVpnConfigured - TRUE if a VPN configuration has been applied to the camera.
 * @param {bool} isVpnRunning - TRUE if the VPN service is actively running.
 */
function updateVPNSection(isAppRunning, isVpnConfigured, isVpnRunning, tunnelIp) {
    let overallStatus = isAppRunning && isVpnRunning && isVpnConfigured && tunnelIp !== '' && tunnelIp !== '\n';
    updateOverallStatusElement(elements.connectionStatusIcon, overallStatus);

    let displayIp = tunnelIp === '' || tunnelIp === '\n' ? '--' : tunnelIp;
    elements.vpnIp.innerHTML = `Tunnel IP Address : ${displayIp}`;

    if (!isAppRunning) {
        // Since the App is not running, we don't need to show any controls
        elements.vpnActions.classList.add('hidden');
        elements.setNewConfigurationButton.classList.add('hidden');

        elements.vpnStatus.innerHTML = 'The Alto Connector ACAP application is not currently running. Start the ACAP app to configure the connection.';
    } else if (!isVpnConfigured) {
        // Since the VPN is not configured, we don't need to show the VPN controls.
        elements.vpnActions.classList.add('hidden');

        elements.vpnStatus.innerHTML = 'Configure an Alto VPN configuration to allow Alto to discover and maintain a connection to this camera.';

        // Display the big Configuration Button.
        elements.setNewConfigurationButton.classList.remove('hidden');
    } else {
        // The App is running and the VPN is configured.
        
        // Display the VPN Action controls
        elements.vpnActions.classList.remove('hidden');

        // Hide the Configuration button and allow the user to access this functionality through a link.
        elements.setNewConfigurationButton.classList.add('hidden');
        elements.vpnStatus.innerHTML = 'An Alto VPN Configuration is already set. <a class="text-blue-600 underline cursor-pointer" onclick="clickFileInput()">Set a new configuration.</a>';

        if(isVpnRunning) {
            elements.vpnState.innerHTML = 'Current Status : RUNNING';
        } else {
            elements.vpnState.innerHTML = 'Current Status : STOPPED';
        }
    }
}

/**
 * Update the ONVIF User section of the DOM based on the new state.
 *
 * @param {bool} isOnvifUserConfigured - TRUE if an ONVIF Administrator user is configured on the camera with the name 'alto'.
 */
function updateOnvifUserSection(isOnvifUserConfigured) {
    updateOverallStatusElement(elements.userConfigStatusIcon, isOnvifUserConfigured);

    if (isOnvifUserConfigured) {
        // Change the section message
        elements.onvifUserMessage.innerHTML = 'An ONVIF Administrator user has been created for Orchid Alto!';
        elements.setOnvifAltoUserButton.innerHTML = 'Reset ONVIF User';
    } else {
        // Change the section message
        elements.onvifUserMessage.innerHTML = 'Create an ONVIF Administrator user to allow Orchid Alto to manage this camera.';
        elements.setOnvifAltoUserButton.innerHTML = 'Create ONVIF User';
    }
}

/**
 * Submit the file selected on the file input to the server.
 */
function submitVPNFile() {
    let file = elements.vpnUpload.files[0];

    if (file === undefined || file === null) {
        console.error('No file to send!');
        return;
    }

    // Use FileReader to perform base64 encoding.
    const fileReader = new FileReader();
    fileReader.readAsDataURL(file);
    fileReader.onerror = () => {
        console.error(`Failed to base64 encode file: ${fileReader.error}`);
    };
    fileReader.onload = () => {
        const encodedPayload = fileReader.result.toString().replace(/^data:(.*,)?/, '');
        let url = new URL(vpnConfigPath, baseUrl);

        fetch(url.toString(),
            {
                method: 'POST',
                body: new URLSearchParams({ 'config': encodedPayload })
            })
            .then(response => {
                if (!response.ok) {
                    console.error(`VPN POST returned failure code: ${response.status} - ${response.statusText}`);
                }

                return updateVPNStatus();
            })
            .catch(error => {
                console.error('Error received!');
            });
    }
}

/**
 * Perform the specified VPN action and update the VPN status.
 *
 * @param {string} action - VPN action [start, stop, restart, status]
 */
function vpnAction(action) {
    fetchVPNControl(action)
        .then(response => response.text())
        .then(text => {
            console.log(`${action} : ${text}`)
            return updateVPNStatus();
        })
        .catch(err => {
            console.error(`Error excecuting VPN Action (${action}) : ${err}`);
        });
}

/**
 * Fetch the logs using the API.
 *
 * @returns {Promise} - A promise that resolves with text logs.
 */
function getLogs() {
    const url = new URL(vpnLogPath, baseUrl);
    return fetch(url.toString())
        .then(response => {
            return response.text();
        });
}

function updateLoop() {
    let promiseSet = [];
    
    if (elements.logDetails.open) {
        promiseSet.push(updateLogs());
    }

    if (elements.connectionDetails.open) {
        promiseSet.push(updateVPNStatus());
    }

    Promise.allSettled(promiseSet).then(() => {
        setTimeout(updateLoop, 1000);
    });
}
/**
 * Update the log display.
 */
function updateLogs() {
    return getLogs().then(logText => {
        elements.serviceLog.innerHTML = logText;
    });    
}


/**
 * Send a SOAP request to the camera.
 *
 * @param {string} body - SOAP request body.
 * @returns {Promise} - A Promise that resolves with the response text on success. Rejects if an error status was returned.
 */
function sendSoapRequest(body) {
    return new Promise((resolve, reject) => {
        let url = new URL(vapixServicesPath, baseUrl);
        let xmlHttp = new XMLHttpRequest(); 
        xmlHttp.open('POST', url.toString(), true);

        xmlHttp.onload = () => {
            if (xmlHttp.status >= 200 && xmlHttp.status < 300) {
                resolve(xmlHttp.responseText);
            } else {
                reject({
                    status: xmlHttp.status,
                    statusText: xmlHttp.statusText
                });
            }
        };

        xmlHttp.onerror = () => {
            reject({
                status: xmlHttp.status,
                statusText: xmlHttp.statusText
            });
        };        

        xmlHttp.setRequestHeader('Content-Type', 'text/xml');
        xmlHttp.send(body);
    });
}

/**
 * Handle onvif user button clicks to create/update the Alto Onvif user.
 */
function setOnvifAltoUser() {
    sendSoapRequest(soapUpdateUserRequest(onvifAltoUserExists, elements.password.value))
        .then(checkOnvifAltoUser)
        .catch(err => {
            console.error(`Error setting user : ${err}`);
        })
        .finally(() => {
            elements.password.value = '';
        });
}

/**
 * Fetch the ONVIF users, check if the Alto user exists, and update the DOM.
 */
function checkOnvifAltoUser() {
    return sendSoapRequest(soapGetUsersRequest)
        .then(responseText => {
            const parser = new DOMParser();
            const xmlDoc = parser.parseFromString(responseText, 'application/xml');
            onvifAltoUserExists = hasValidOnvifAltoUser(xmlDoc);
            updateOnvifUserSection(onvifAltoUserExists);
        });
}

/**
 * Determine if the GetUSers repsonse contains the Alto ONVIF user.
 *
 * @param {XMLDocument} xmlDoc - The XML document containing the GetUsers SOAP response.
 * @returns {bool} - TRUE if the response contains an Alto administrator user.
 */
function hasValidOnvifAltoUser(xmlDoc) {
    let userElements = xmlDoc.getElementsByTagName('tds:User');
    
    if (userElements.length < 1) return false;

    for (let i = 0; i < userElements.length; i++) {
        let children = userElements[i].children;
        let isAltoUser = false;
        let isAdmin = false;
        for (let j = 0; j < children.length; j++) {
            if (children[j].tagName === 'tt:Username' && children[j].innerHTML === 'alto') {
                isAltoUser = true;
            } 

            if (children[j].tagName === 'tt:UserLevel' && children[j].innerHTML === 'Administrator') {
                isAdmin = true;
            }

            if (isAltoUser && isAdmin) return true;
        }
    }

    return false;
}

/**
 * Retrieve Elements from the DOM.
 *
 * @returns {Object} Object containing the HTML Elements.
 */
function getDOMElements() {
    const elements = {
        vpnStatus: document.getElementById('vpnStatus'),
        vpnState: document.getElementById('vpnState'),
        vpnUpload: document.getElementById('vpnUpload'),
        vpnIp: document.getElementById('vpnIp'),
        logDetails: document.getElementById('logDetails'),
        connectionDetails: document.getElementById('connectionDetails'),
        serviceLog: document.getElementById('serviceLog'),
        onvifUserMessage: document.getElementById('onvifUserMessage'),
        setOnvifAltoUserButton: document.getElementById('setOnvifAltoUserButton'),
        password: document.getElementById('password'),
        userConfigStatusIcon: document.getElementById('userConfigStatusIcon'),
        connectionStatusIcon: document.getElementById('connectionStatusIcon'),
        vpnActions: document.getElementById('vpnActions'),
        setNewConfigurationButton: document.getElementById('setNewConfigurationButton')
    }
    
    // Check that all of the expected DOM elements exist.
    for (const [key, value] of Object.entries(elements)) {
        if (!value) alert(`DOM Element undefined : ${key}`);
    }

    return elements;
}

updateVPNStatus();
updateLoop();
checkOnvifAltoUser().catch(err => console.error(err));
